@@ -99,6 +99,7 @@ class BinaryReader {
   Result ReadS32Leb128(uint32_t* out_value, const char* desc) WABT_WARN_UNUSED;
   Result ReadS64Leb128(uint64_t* out_value, const char* desc) WABT_WARN_UNUSED;
   Result ReadType(Type* out_value, const char* desc) WABT_WARN_UNUSED;
+  Result ReadRefType(Type* out_value, const char* desc) WABT_WARN_UNUSED;
   Result ReadExternalKind(ExternalKind* out_value,
                           const char* desc) WABT_WARN_UNUSED;
   Result ReadStr(string_view* out_str, const char* desc) WABT_WARN_UNUSED;
@@ -107,6 +108,7 @@ class BinaryReader {
                    const char* desc) WABT_WARN_UNUSED;
   Result ReadIndex(Index* index, const char* desc) WABT_WARN_UNUSED;
   Result ReadOffset(Offset* offset, const char* desc) WABT_WARN_UNUSED;
+  Result ReadAlignment(uint32_t* align_log2, const char* desc) WABT_WARN_UNUSED;
   Result ReadCount(Index* index, const char* desc) WABT_WARN_UNUSED;
   Result ReadField(TypeMut* out_value) WABT_WARN_UNUSED;
 
@@ -302,6 +304,14 @@ Result BinaryReader::ReadType(Type* out_value, const char* desc) {
   return Result::Ok;
 }
 
+Result BinaryReader::ReadRefType(Type* out_value, const char* desc) {
+  uint32_t type = 0;
+  CHECK_RESULT(ReadS32Leb128(&type, desc));
+  *out_value = static_cast<Type>(type);
+  ERROR_UNLESS(out_value->IsRef(), "%s must be a reference type", desc);
+  return Result::Ok;
+}
+
 Result BinaryReader::ReadExternalKind(ExternalKind* out_value,
                                       const char* desc) {
   uint8_t value = 0;
@@ -357,6 +367,17 @@ Result BinaryReader::ReadOffset(Offset* offset, const char* desc) {
   return Result::Ok;
 }
 
+Result BinaryReader::ReadAlignment(uint32_t* alignment_log2, const char* desc) {
+  uint32_t value;
+  CHECK_RESULT(ReadU32Leb128(&value, desc));
+  if (value >= 32) {
+    PrintError("invalid %s: %u", desc, value);
+    return Result::Error;
+  }
+  *alignment_log2 = value;
+  return Result::Ok;
+}
+
 Result BinaryReader::ReadCount(Index* count, const char* desc) {
   CHECK_RESULT(ReadIndex(count, desc));
 
@@ -401,12 +422,11 @@ bool BinaryReader::IsConcreteType(Type type) {
     case Type::V128:
       return options_.features.simd_enabled();
 
-    case Type::Funcref:
-    case Type::Anyref:
-    case Type::Nullref:
+    case Type::FuncRef:
+    case Type::ExternRef:
       return options_.features.reference_types_enabled();
 
-    case Type::Exnref:
+    case Type::ExnRef:
       return options_.features.exceptions_enabled();
 
     default:
@@ -483,9 +503,12 @@ Result BinaryReader::ReadInitExpr(Index index, bool require_i32) {
       break;
     }
 
-    case Opcode::RefNull:
-      CALLBACK(OnInitExprRefNull, index);
+    case Opcode::RefNull: {
+      Type type;
+      CHECK_RESULT(ReadRefType(&type, "ref.null type"));
+      CALLBACK(OnInitExprRefNull, index, type);
       break;
+    }
 
     case Opcode::RefFunc: {
       Index func_index;
@@ -514,9 +537,7 @@ Result BinaryReader::ReadInitExpr(Index index, bool require_i32) {
 }
 
 Result BinaryReader::ReadTable(Type* out_elem_type, Limits* out_elem_limits) {
-  CHECK_RESULT(ReadType(out_elem_type, "table elem type"));
-  ERROR_UNLESS(out_elem_type->IsRef(),
-               "table elem type must be a reference type");
+  CHECK_RESULT(ReadRefType(out_elem_type, "table elem type"));
 
   uint32_t flags;
   uint32_t initial;
@@ -634,7 +655,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       }
 
       case Opcode::Select:
-        CALLBACK(OnSelectExpr, Type::Any);
+        CALLBACK(OnSelectExpr, Type::Void);
         CALLBACK0(OnOpcodeBare);
         break;
 
@@ -853,7 +874,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::I64X2Load32X2S:
       case Opcode::I64X2Load32X2U: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "load alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "load alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "load offset"));
         CALLBACK(OnLoadExpr, opcode, alignment_log2, offset);
@@ -872,7 +893,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::F64Store:
       case Opcode::V128Store: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "store alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "store alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "store offset"));
 
@@ -1188,7 +1209,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::V32X4LoadSplat:
       case Opcode::V64X2LoadSplat: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "load alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "load alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "load offset"));
 
@@ -1294,7 +1315,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
 
       case Opcode::AtomicNotify: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "load alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "load alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "load offset"));
 
@@ -1306,7 +1327,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::I32AtomicWait:
       case Opcode::I64AtomicWait: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "load alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "load alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "load offset"));
 
@@ -1333,7 +1354,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::I32AtomicLoad:
       case Opcode::I64AtomicLoad: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "load alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "load alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "load offset"));
 
@@ -1350,7 +1371,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::I32AtomicStore:
       case Opcode::I64AtomicStore: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "store alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "store alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "store offset"));
 
@@ -1402,7 +1423,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::I64AtomicRmw16XchgU:
       case Opcode::I64AtomicRmw32XchgU: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "memory alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "memory alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "memory offset"));
 
@@ -1419,7 +1440,7 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       case Opcode::I64AtomicRmw16CmpxchgU:
       case Opcode::I64AtomicRmw32CmpxchgU: {
         uint32_t alignment_log2;
-        CHECK_RESULT(ReadU32Leb128(&alignment_log2, "memory alignment"));
+        CHECK_RESULT(ReadAlignment(&alignment_log2, "memory alignment"));
         Address offset;
         CHECK_RESULT(ReadU32Leb128(&offset, "memory offset"));
 
@@ -1545,14 +1566,18 @@ Result BinaryReader::ReadFunctionBody(Offset end_offset) {
       }
 
       case Opcode::RefNull: {
-        CALLBACK(OnRefNullExpr);
-        CALLBACK0(OnOpcodeBare);
+        Type type;
+        CHECK_RESULT(ReadRefType(&type, "ref.null type"));
+        CALLBACK(OnRefNullExpr, type);
+        CALLBACK(OnOpcodeType, type);
         break;
       }
 
       case Opcode::RefIsNull: {
-        CALLBACK(OnRefIsNullExpr);
-        CALLBACK0(OnOpcodeBare);
+        Type type;
+        CHECK_RESULT(ReadRefType(&type, "ref.is_null type"));
+        CALLBACK(OnRefIsNullExpr, type);
+        CALLBACK(OnOpcodeType, type);
         break;
       }
 
@@ -1817,7 +1842,7 @@ Result BinaryReader::ReadLinkingSection(Offset section_size) {
           uint32_t alignment_log2;
           uint32_t flags;
           CHECK_RESULT(ReadStr(&name, "segment name"));
-          CHECK_RESULT(ReadU32Leb128(&alignment_log2, "segment alignment"));
+          CHECK_RESULT(ReadAlignment(&alignment_log2, "segment alignment"));
           CHECK_RESULT(ReadU32Leb128(&flags, "segment flags"));
           CALLBACK(OnSegmentInfo, i, name, alignment_log2, flags);
         }
@@ -2184,7 +2209,7 @@ Result BinaryReader::ReadElemSection(Offset section_size) {
     if ((flags & (SegPassive | SegExplicitIndex)) == SegExplicitIndex) {
       CHECK_RESULT(ReadIndex(&table_index, "elem segment table index"));
     }
-    Type elem_type = Type::Funcref;
+    Type elem_type = Type::FuncRef;
 
     CALLBACK(BeginElemSegment, i, table_index, flags);
 
@@ -2197,17 +2222,14 @@ Result BinaryReader::ReadElemSection(Offset section_size) {
     // For backwards compat we support not declaring the element kind.
     if (flags & (SegPassive | SegExplicitIndex)) {
       if (flags & SegUseElemExprs) {
-        CHECK_RESULT(ReadType(&elem_type, "table elem type"));
-        ERROR_UNLESS(elem_type.IsRef(),
-                     "segment elem expr type must be a reference type (got %s)",
-                     elem_type.GetName());
+        CHECK_RESULT(ReadRefType(&elem_type, "table elem type"));
       } else {
         ExternalKind kind;
         CHECK_RESULT(ReadExternalKind(&kind, "export kind"));
         ERROR_UNLESS(kind == ExternalKind::Func,
                      "segment elem type must be func (%s)",
                      elem_type.GetName());
-        elem_type = Type::Funcref;
+        elem_type = Type::FuncRef;
       }
     }
 
@@ -2222,7 +2244,9 @@ Result BinaryReader::ReadElemSection(Offset section_size) {
         Opcode opcode;
         CHECK_RESULT(ReadOpcode(&opcode, "elem expr opcode"));
         if (opcode == Opcode::RefNull) {
-          CALLBACK(OnElemSegmentElemExpr_RefNull, i);
+          Type type;
+          CHECK_RESULT(ReadRefType(&type, "elem expr ref.null type"));
+          CALLBACK(OnElemSegmentElemExpr_RefNull, i, type);
         } else if (opcode == Opcode::RefFunc) {
           Index func_index;
           CHECK_RESULT(ReadIndex(&func_index, "elem expr func index"));
