@@ -433,12 +433,12 @@ Result Validator::OnRefFuncExpr(RefFuncExpr* expr) {
 }
 
 Result Validator::OnRefNullExpr(RefNullExpr* expr) {
-  result_ |= validator_.OnRefNull(expr->loc);
+  result_ |= validator_.OnRefNull(expr->loc, expr->type);
   return Result::Ok;
 }
 
 Result Validator::OnRefIsNullExpr(RefIsNullExpr* expr) {
-  result_ |= validator_.OnRefIsNull(expr->loc);
+  result_ |= validator_.OnRefIsNull(expr->loc, expr->type);
   return Result::Ok;
 }
 
@@ -464,8 +464,13 @@ Result Validator::OnReturnCallIndirectExpr(ReturnCallIndirectExpr* expr) {
 }
 
 Result Validator::OnSelectExpr(SelectExpr* expr) {
-  assert(expr->result_type.size());
-  result_ |= validator_.OnSelect(expr->loc, expr->result_type[0]);
+  Type result_type;
+  if (expr->result_type.empty()) {
+    result_type = Type::Void;
+  } else {
+    result_type = expr->result_type[0];
+  }
+  result_ |= validator_.OnSelect(expr->loc, result_type);
   // TODO: Existing behavior fails when select fails.
 #if 0
   return Result::Ok;
@@ -715,7 +720,8 @@ Result Validator::CheckModule() {
             break;
 
           case ExprType::RefNull:
-            result_ |= validator_.OnGlobalInitExpr_RefNull(expr->loc);
+            result_ |= validator_.OnGlobalInitExpr_RefNull(
+                expr->loc, cast<RefNullExpr>(expr)->type);
             break;
 
           default:
@@ -789,7 +795,8 @@ Result Validator::CheckModule() {
         switch (elem_expr.kind) {
           case ElemExprKind::RefNull:
             // TODO: better location?
-            result_ |= validator_.OnElemSegmentElemExpr_RefNull(field.loc);
+            result_ |= validator_.OnElemSegmentElemExpr_RefNull(field.loc,
+                                                                elem_expr.type);
             break;
 
           case ElemExprKind::RefFunc:
