@@ -229,7 +229,9 @@ bool IsCommand(TokenTypePair pair) {
     case TokenType::AssertUnlinkable:
     case TokenType::Get:
     case TokenType::Invoke:
+    case TokenType::Input:
     case TokenType::Module:
+    case TokenType::Output:
     case TokenType::Register:
       return true;
     default:
@@ -241,7 +243,7 @@ bool IsEmptySignature(const FuncSignature& sig) {
   return sig.result_types.empty() && sig.param_types.empty();
 }
 
-void ResolveFuncTypeWithEmptySignature(const Module& module,
+bool ResolveFuncTypeWithEmptySignature(const Module& module,
                                        FuncDeclaration* decl) {
   // Resolve func type variables where the signature was not specified
   // explicitly, e.g.: (func (type 1) ...)
@@ -249,9 +251,10 @@ void ResolveFuncTypeWithEmptySignature(const Module& module,
     const FuncType* func_type = module.GetFuncType(decl->type_var);
     if (func_type) {
       decl->sig = func_type->sig;
+      return true;
     }
   }
-
+  return false;
 }
 
 void ResolveImplicitlyDefinedFunctionType(const Location& loc,
@@ -417,14 +420,28 @@ Result ResolveFuncTypes(Module* module, Errors* errors) {
       continue;
     }
 
+    bool has_func_type_and_empty_signature = false;
+
     if (decl) {
-      ResolveFuncTypeWithEmptySignature(*module, decl);
+      has_func_type_and_empty_signature =
+          ResolveFuncTypeWithEmptySignature(*module, decl);
       ResolveImplicitlyDefinedFunctionType(field.loc, module, *decl);
       result |=
           CheckFuncTypeVarMatchesExplicit(field.loc, *module, *decl, errors);
     }
 
     if (func) {
+      if (has_func_type_and_empty_signature) {
+        // The call to ResolveFuncTypeWithEmptySignature may have updated the
+        // function signature so there are parameters. Since parameters and
+        // local variables share the same index space, we need to increment the
+        // local indexes bound to a given name by the number of parameters in
+        // the function.
+        for (auto& pair: func->bindings) {
+          pair.second.index += func->GetNumParams();
+        }
+      }
+
       ResolveFuncTypesExprVisitorDelegate delegate(module, errors);
       ExprVisitor visitor(&delegate);
       result |= visitor.VisitFunc(func);
@@ -731,7 +748,9 @@ bool WastParser::ParseElemExprOpt(ElemExpr* out_elem_expr) {
           options_->features.reference_types_enabled())) {
       Error(loc, "ref.null not allowed");
     }
-    *out_elem_expr = ElemExpr();
+    Type type;
+    CHECK_RESULT(ParseRefKind(&type));
+    *out_elem_expr = ElemExpr(type);
   } else if (Match(TokenType::RefFunc)) {
     Var var;
     CHECK_RESULT(ParseVar(&var));
@@ -768,7 +787,7 @@ bool WastParser::ParseElemExprVarListOpt(ElemExprVector* out_list) {
 Result WastParser::ParseValueType(Type* out_type) {
   WABT_TRACE(ParseValueType);
   if (!PeekMatch(TokenType::ValueType)) {
-    return ErrorExpected({"i32", "i64", "f32", "f64", "v128", "anyref"});
+    return ErrorExpected({"i32", "i64", "f32", "f64", "v128", "externref"});
   }
 
   Token token = Consume();
@@ -778,12 +797,11 @@ Result WastParser::ParseValueType(Type* out_type) {
     case Type::V128:
       is_enabled = options_->features.simd_enabled();
       break;
-    case Type::Anyref:
-    case Type::Funcref:
-    case Type::Hostref:
+    case Type::FuncRef:
+    case Type::ExternRef:
       is_enabled = options_->features.reference_types_enabled();
       break;
-    case Type::Exnref:
+    case Type::ExnRef:
       is_enabled = options_->features.exceptions_enabled();
       break;
     default:
@@ -808,15 +826,37 @@ Result WastParser::ParseValueTypeList(TypeVector* out_type_list) {
   return Result::Ok;
 }
 
-Result WastParser::ParseRefType(Type* out_type) {
-  WABT_TRACE(ParseRefType);
-  if (!PeekMatch(TokenType::ValueType)) {
-    return ErrorExpected({"funcref", "anyref", "nullref", "exnref"});
+Result WastParser::ParseRefKind(Type* out_type) {
+  WABT_TRACE(ParseRefKind);
+  if (!IsTokenTypeRefKind(Peek())) {
+    return ErrorExpected({"func", "extern", "exn"});
   }
 
   Token token = Consume();
   Type type = token.type();
-  if (type == Type::Anyref && !options_->features.reference_types_enabled()) {
+
+  if ((type == Type::ExternRef &&
+       !options_->features.reference_types_enabled()) ||
+      ((type == Type::Struct || type == Type::Array) &&
+       !options_->features.gc_enabled())) {
+    Error(token.loc, "value type not allowed: %s", type.GetName());
+    return Result::Error;
+  }
+
+  *out_type = type;
+  return Result::Ok;
+}
+
+Result WastParser::ParseRefType(Type* out_type) {
+  WABT_TRACE(ParseRefType);
+  if (!PeekMatch(TokenType::ValueType)) {
+    return ErrorExpected({"funcref", "externref", "exnref"});
+  }
+
+  Token token = Consume();
+  Type type = token.type();
+  if (type == Type::ExternRef &&
+      !options_->features.reference_types_enabled()) {
     Error(token.loc, "value type not allowed: %s", type.GetName());
     return Result::Error;
   }
@@ -833,7 +873,8 @@ bool WastParser::ParseRefTypeOpt(Type* out_type) {
 
   Token token = Consume();
   Type type = token.type();
-  if (type == Type::Anyref && !options_->features.reference_types_enabled()) {
+  if (type == Type::ExternRef &&
+      !options_->features.reference_types_enabled()) {
     return false;
   }
 
@@ -1107,7 +1148,7 @@ Result WastParser::ParseElemModuleField(Module* module) {
   if (ParseRefTypeOpt(&field->elem_segment.elem_type)) {
     ParseElemExprListOpt(&field->elem_segment.elem_exprs);
   } else {
-    field->elem_segment.elem_type = Type::Funcref;
+    field->elem_segment.elem_type = Type::FuncRef;
     if (PeekMatch(TokenType::Func)) {
       EXPECT(Func);
     }
@@ -1712,8 +1753,6 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
           MatchLpar(TokenType::Result)) {
         CHECK_RESULT(ParseValueTypeList(&result));
         EXPECT(Rpar);
-      } else {
-        result.push_back(Type::Any);
       }
       out_expr->reset(new SelectExpr(result, loc));
       break;
@@ -1949,15 +1988,21 @@ Result WastParser::ParsePlainInstr(std::unique_ptr<Expr>* out_expr) {
       CHECK_RESULT(ParsePlainInstrVar<RefFuncExpr>(loc, out_expr));
       break;
 
-    case TokenType::RefNull:
+    case TokenType::RefNull: {
       ErrorUnlessOpcodeEnabled(Consume());
-      out_expr->reset(new RefNullExpr(loc));
+      Type type;
+      CHECK_RESULT(ParseRefKind(&type));
+      out_expr->reset(new RefNullExpr(type, loc));
       break;
+    }
 
-    case TokenType::RefIsNull:
+    case TokenType::RefIsNull: {
       ErrorUnlessOpcodeEnabled(Consume());
-      out_expr->reset(new RefIsNullExpr(loc));
+      Type type;
+      CHECK_RESULT(ParseRefKind(&type));
+      out_expr->reset(new RefIsNullExpr(type, loc));
       break;
+    }
 
     case TokenType::Throw:
       ErrorUnlessOpcodeEnabled(Consume());
@@ -2360,11 +2405,11 @@ Result WastParser::ParseConst(Const* const_, ConstType const_type) {
   return Result::Ok;
 }
 
-Result WastParser::ParseHostRef(Const* const_) {
-  WABT_TRACE(ParseHostRef);
+Result WastParser::ParseExternref(Const* const_) {
+  WABT_TRACE(ParseExternref);
   Token token = Consume();
   if (!options_->features.reference_types_enabled()) {
-    Error(token.loc, "hostref not allowed");
+    Error(token.loc, "externref not allowed");
     return Result::Error;
   }
 
@@ -2391,7 +2436,7 @@ Result WastParser::ParseHostRef(Const* const_) {
   uint64_t ref_bits;
   Result result = ParseInt64(s, end, &ref_bits, ParseIntType::UnsignedOnly);
 
-  const_->set_hostref(static_cast<uintptr_t>(ref_bits));
+  const_->set_externref(static_cast<uintptr_t>(ref_bits));
 
   if (Failed(result)) {
     Error(const_->loc, "invalid literal \"" PRIstringview "\"",
@@ -2406,7 +2451,7 @@ Result WastParser::ParseHostRef(Const* const_) {
 Result WastParser::ParseConstList(ConstVector* consts, ConstType type) {
   WABT_TRACE(ParseConstList);
   while (PeekMatchLpar(TokenType::Const) || PeekMatchLpar(TokenType::RefNull) ||
-         PeekMatchLpar(TokenType::RefHost) ||
+         PeekMatchLpar(TokenType::RefExtern) ||
          PeekMatchLpar(TokenType::RefFunc)) {
     Consume();
     Const const_;
@@ -2416,9 +2461,11 @@ Result WastParser::ParseConstList(ConstVector* consts, ConstType type) {
         break;
       case TokenType::RefNull: {
         auto token = Consume();
+        Type type;
+        CHECK_RESULT(ParseRefKind(&type));
         ErrorUnlessOpcodeEnabled(token);
         const_.loc = GetLocation();
-        const_.set_nullref();
+        const_.set_null(type);
         break;
       }
       case TokenType::RefFunc: {
@@ -2428,8 +2475,8 @@ Result WastParser::ParseConstList(ConstVector* consts, ConstType type) {
         const_.set_funcref();
         break;
       }
-      case TokenType::RefHost:
-        CHECK_RESULT(ParseHostRef(&const_));
+      case TokenType::RefExtern:
+        CHECK_RESULT(ParseExternref(&const_));
         break;
       default:
         assert(!"unreachable");
@@ -2734,6 +2781,12 @@ Result WastParser::ParseCommand(Script* script, CommandPtr* out_command) {
     case TokenType::Register:
       return ParseRegisterCommand(out_command);
 
+    case TokenType::Input:
+      return ParseInputCommand(out_command);
+
+    case TokenType::Output:
+      return ParseOutputCommand(out_command);
+
     default:
       assert(!"ParseCommand should only be called when IsCommand() is true");
       return Result::Error;
@@ -2876,6 +2929,40 @@ Result WastParser::ParseRegisterCommand(CommandPtr* out_command) {
   return Result::Ok;
 }
 
+Result WastParser::ParseInputCommand(CommandPtr*) {
+  // Parse the input command, but always fail since this command is not
+  // actually supported.
+  WABT_TRACE(ParseInputCommand);
+  EXPECT(Lpar);
+  Location loc = GetLocation();
+  EXPECT(Input);
+  Error(loc, "input command is not supported");
+  Var var;
+  std::string text;
+  ParseVarOpt(&var);
+  CHECK_RESULT(ParseQuotedText(&text));
+  EXPECT(Rpar);
+  return Result::Error;
+}
+
+Result WastParser::ParseOutputCommand(CommandPtr*) {
+  // Parse the output command, but always fail since this command is not
+  // actually supported.
+  WABT_TRACE(ParseOutputCommand);
+  EXPECT(Lpar);
+  Location loc = GetLocation();
+  EXPECT(Output);
+  Error(loc, "output command is not supported");
+  Var var;
+  std::string text;
+  ParseVarOpt(&var);
+  if (Peek() == TokenType::Text) {
+    CHECK_RESULT(ParseQuotedText(&text));
+  }
+  EXPECT(Rpar);
+  return Result::Error;
+}
+
 Result WastParser::ParseAction(ActionPtr* out_action) {
   WABT_TRACE(ParseAction);
   EXPECT(Lpar);
