@@ -121,7 +121,9 @@ class WatWriter : ModuleContext {
   void WriteQuotedData(const void* data, size_t length);
   void WriteQuotedString(string_view str, NextChar next_char);
   void WriteVar(const Var& var, NextChar next_char);
+  void WriteVarUnlessZero(const Var& var, NextChar next_char);
   void WriteBrVar(const Var& var, NextChar next_char);
+  void WriteRefKind(Type type, NextChar next_char);
   void WriteType(Type type, NextChar next_char);
   void WriteTypes(const TypeVector& types, const char* name);
   void WriteFuncSigSpace(const FuncSignature& func_sig);
@@ -363,6 +365,16 @@ void WatWriter::WriteVar(const Var& var, NextChar next_char) {
   }
 }
 
+bool VarIsZero(const Var& var) {
+  return var.is_index() && var.index() == 0;
+}
+
+void WatWriter::WriteVarUnlessZero(const Var& var, NextChar next_char) {
+  if (!VarIsZero(var)) {
+    WriteVar(var, next_char);
+  }
+}
+
 void WatWriter::WriteBrVar(const Var& var, NextChar next_char) {
   if (var.is_index()) {
     if (var.index() < GetLabelStackSize()) {
@@ -377,6 +389,10 @@ void WatWriter::WriteBrVar(const Var& var, NextChar next_char) {
   }
 }
 
+void WatWriter::WriteRefKind(Type type, NextChar next_char) {
+  WritePuts(type.GetRefKindName(), next_char);
+}
+
 void WatWriter::WriteType(Type type, NextChar next_char) {
   const char* type_name = type.GetName();
   assert(type_name);
@@ -615,9 +631,7 @@ Result WatWriter::ExprVisitorDelegate::OnCallExpr(CallExpr* expr) {
 Result WatWriter::ExprVisitorDelegate::OnCallIndirectExpr(
     CallIndirectExpr* expr) {
   writer_->WritePutsSpace(Opcode::CallIndirect_Opcode.GetName());
-  if (!expr->table.is_index() || expr->table.index() != 0) {
-    writer_->WriteVar(expr->table, NextChar::Space);
-  }
+  writer_->WriteVarUnlessZero(expr->table, NextChar::Space);
   writer_->WriteOpenSpace("type");
   writer_->WriteVar(expr->decl.type_var, NextChar::Newline);
   writer_->WriteCloseNewline();
@@ -744,7 +758,12 @@ Result WatWriter::ExprVisitorDelegate::OnMemoryInitExpr(MemoryInitExpr* expr) {
 }
 
 Result WatWriter::ExprVisitorDelegate::OnTableCopyExpr(TableCopyExpr* expr) {
-  writer_->WritePutsNewline(Opcode::TableCopy_Opcode.GetName());
+  writer_->WritePutsSpace(Opcode::TableCopy_Opcode.GetName());
+  if (!(VarIsZero(expr->dst_table) && VarIsZero(expr->src_table))) {
+    writer_->WriteVar(expr->dst_table, NextChar::Space);
+    writer_->WriteVar(expr->src_table, NextChar::Space);
+  }
+  writer_->WriteNewline(NO_FORCE_NEWLINE);
   return Result::Ok;
 }
 
@@ -756,6 +775,7 @@ Result WatWriter::ExprVisitorDelegate::OnElemDropExpr(ElemDropExpr* expr) {
 
 Result WatWriter::ExprVisitorDelegate::OnTableInitExpr(TableInitExpr* expr) {
   writer_->WritePutsSpace(Opcode::TableInit_Opcode.GetName());
+  writer_->WriteVarUnlessZero(expr->table_index, NextChar::Space);
   writer_->WriteVar(expr->segment_index, NextChar::Newline);
   return Result::Ok;
 }
@@ -797,12 +817,14 @@ Result WatWriter::ExprVisitorDelegate::OnRefFuncExpr(RefFuncExpr* expr) {
 }
 
 Result WatWriter::ExprVisitorDelegate::OnRefNullExpr(RefNullExpr* expr) {
-  writer_->WritePutsNewline(Opcode::RefNull_Opcode.GetName());
+  writer_->WritePutsSpace(Opcode::RefNull_Opcode.GetName());
+  writer_->WriteRefKind(expr->type, NextChar::Newline);
   return Result::Ok;
 }
 
 Result WatWriter::ExprVisitorDelegate::OnRefIsNullExpr(RefIsNullExpr* expr) {
-  writer_->WritePutsNewline(Opcode::RefIsNull_Opcode.GetName());
+  writer_->WritePutsSpace(Opcode::RefIsNull_Opcode.GetName());
+  writer_->WriteRefKind(expr->type, NextChar::Newline);
   return Result::Ok;
 }
 
@@ -832,7 +854,11 @@ Result WatWriter::ExprVisitorDelegate::OnReturnCallIndirectExpr(
 }
 
 Result WatWriter::ExprVisitorDelegate::OnSelectExpr(SelectExpr* expr) {
-  writer_->WritePutsNewline(Opcode::Select_Opcode.GetName());
+  writer_->WritePutsSpace(Opcode::Select_Opcode.GetName());
+  if (!expr->result_type.empty()) {
+    writer_->WriteTypes(expr->result_type, "result");
+  }
+  writer_->WriteNewline(NO_FORCE_NEWLINE);
   return Result::Ok;
 }
 
@@ -1268,18 +1294,24 @@ void WatWriter::WriteElemSegment(const ElemSegment& segment) {
 
   uint8_t flags = segment.GetFlags(&module);
 
+  if ((flags & (SegPassive | SegExplicitIndex)) == SegExplicitIndex) {
+    WriteOpenSpace("table");
+    WriteVar(segment.table_var, NextChar::Space);
+    WriteCloseSpace();
+  }
+
   if (!(flags & SegPassive)) {
     WriteInitExpr(segment.offset);
   }
 
-  if (flags == SegDeclared) {
+  if ((flags & SegDeclared) == SegDeclared) {
     WritePuts("declare", NextChar::Space);
   }
 
   if (flags & SegUseElemExprs) {
     WriteType(segment.elem_type, NextChar::Space);
   } else {
-    assert(segment.elem_type == Type::Funcref);
+    assert(segment.elem_type == Type::FuncRef);
     WritePuts("func", NextChar::Space);
   }
 
@@ -1287,6 +1319,7 @@ void WatWriter::WriteElemSegment(const ElemSegment& segment) {
     if (flags & SegUseElemExprs) {
       if (expr.kind == ElemExprKind::RefNull) {
         WriteOpenSpace("ref.null");
+        WriteRefKind(expr.type, NextChar::Space);
         WriteCloseSpace();
       } else {
         WriteOpenSpace("ref.func");
